/*
 * C
 *
 * Copyright 2018-2019 MicroEJ Corp. All rights reserved.
 * MicroEJ Corp. PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */

#include <stdint.h>
#include <wolfssl/wolfcrypt/asn.h>
#include <wolfssl/wolfcrypt/rsa.h>
#include <wolfssl/wolfcrypt/sha256.h>
#include <wolfssl/wolfcrypt/error-crypt.h>
#include <sni.h>

/**
 * @file
 * @brief RS256 implementation over Wolfcrypt signature.
 * @author @CCO_AUTHOR@
 * @version @CCO_VERSION@
 * @date @CCO_DATE@
 */

#ifdef __cplusplus
	extern "C" {
#endif

/**
 * @brief rsa key, only one key is managed right now.
 */
static RsaKey key;

/**
 * @brief random number generator.
 */
static RNG rng;

/**
 * @brief Initializes the memory shared by Wolfssl and Wolfscrypt.
 */
extern void LLNET_SSL_MEM_initialize();

/**
 * @brief The offset of the rsa private key infos in a pkcs8 der encoded bytes.
 */
#define PKCS8_RSA_PRIVATE_KEY_OFFSET 26

/**
 * @brief Initialize the rsa key and the rng.
 *
 * @param[in] privateKey the PKCS8 DER encoded private key.
 * @param[in] privateKeyLength the length of the private key.
 * @return the max signature size of the current rsa key.
 */
int32_t Java_com_microej_signature_RS256Signer_initialize(const uint8_t* privateKey, const uint32_t privateKeyLength) {
  int ret;
  word32 idx = PKCS8_RSA_PRIVATE_KEY_OFFSET;

  LLNET_SSL_MEM_initialize();

  if ((ret =  wc_InitRsaKey(&key, NULL)) != 0) {
    return ret;
  } else if ((ret = wc_RsaPrivateKeyDecode(privateKey, &idx, &key, privateKeyLength)) != 0) {
      return ret;
  } else if ((ret = wc_InitRng(&rng)) != 0) {
      return ret;
  }

  ret = wc_RsaEncryptSize(&key);
  if (ret < 0) {
  }

  return ret;
}

/**
 * @brief perform an rs256 signature on the given data.
 *
 * @param[in] data the data to sign.
 * @param[in] dataLength the length of the data to sign.
 * @param[in] signature the array where to store the signature.
 * @param[in,out] signatureLength the length of the signature.
 * @return the length of the signature.
 */
int32_t Java_com_microej_signature_RS256Signer_sign___3BI_3BI(const uint8_t* data, uint32_t dataLength, uint8_t* signature, uint32_t signatureLength) {
  uint8_t hash_data[SHA256_DIGEST_SIZE];
  uint8_t digest_buf [MAX_ENCODED_SIG_SZ];
  int ret = 0;
  
  if ((ret = wc_Sha256Hash(data, dataLength, hash_data)) < 0) {
    return ret;
  } else if ((ret = wc_EncodeSignature(digest_buf, hash_data, SHA256_DIGEST_SIZE, SHA256h)) < 0) {
    return ret;
  } else if ((ret = wc_RsaSSL_Sign(digest_buf, ret, signature, signatureLength, &key, &rng)) < 0) {
    return ret;
  }
  
  return ret;
}

int32_t Java_com_microej_signature_RS256Signer_fillErrorMessage(int32_t errorCode, int8_t* buffer, int32_t bufferLength) {
  wc_ErrorString(errorCode, (char*) buffer);
  return strlen((const char*) buffer);
}

#ifdef __cplusplus
	}
#endif
