/*
 * C
 *
 * Copyright 2016-2019 MicroEJ Corp. All rights reserved.
 * For demonstration purpose only.
 * MicroEJ Corp. PROPRIETARY. Use is subject to license terms.
 */
#include <LLNET_SSL_CONTEXT_impl.h>
#include <LLNET_SSL_CONSTANTS.h>
#include <LLNET_SSL_ERRORS.h>
#include <LLNET_SSL_verifyCallback.h>
#include <wolfssl/ssl.h>
#include <stdlib.h>
#include <wolfssl/internal.h>
#include <wolfssl/error-ssl.h>
#ifdef __cplusplus
	extern "C" {
#endif

/*external functions*/
extern int32_t LLNET_SSL_TranslateReturnCode(int32_t wolfSSL_error);
extern int ConfirmNameConstraints(Signer* signer, DecodedCert* cert);
extern int ConfirmSignature(const byte* buf, word32 bufSz,
    const byte* key, word32 keySz, word32 keyOID,
    const byte* sig, word32 sigSz, word32 sigOID,
    void* heap);
		

/*Return the trusted CA if found, otherwise NULL*/
static Signer* LLNET_SSL_VERIFY_getTrustedCA(WOLFSSL_CTX* ctx, DecodedCert* cert)
{	
		if(cert != NULL){
				Signer* ca = NULL;
				#ifndef NO_SKID
						ca = GetCA(ctx->cm, cert->extSubjKeyId);
						if (ca == NULL){
								ca = GetCAByName(ctx->cm, cert->subjectHash);
						}					
				#else /* NO_SKID */
						ca = GetCA(ctx->cm, cert->subjectHash);
				#endif /* NO SKID */
				if(ca != NULL){
					//compare names and public keys
					 if(ca->pubKeySize == cert->pubKeySize 
						 && XMEMCMP(ca->publicKey, cert->publicKey, ca->pubKeySize) == 0
						 && ca->nameLen == cert->subjectCNLen
						 && XSTRNCMP(ca->name, cert->subjectCN, ca->nameLen) == 0){
							 return ca;
					 }
				}
		}

    return NULL;
}

/* Certificate verification callback */
int32_t LLNET_SSL_VERIFY_verifyCallback(int32_t previousVerifySuccess, WOLFSSL_X509_STORE_CTX* store){
	struct VERIFY_CALLBACK_CTX * cbCtx;
	WOLFSSL_CTX* ctx;
	WOLFSSL* ssl;
	int32_t chainLength;
	
#ifdef SESSION_CERTS	
	//request discard session certs
	store->discardSessionCerts = 1;
#endif	
	
	if(previousVerifySuccess){
			//ok when the previous verification doest not fail.
			return 1; 
	}
	
	cbCtx = (struct VERIFY_CALLBACK_CTX *) store->userCtx;
	if(cbCtx == NULL){
		return 0; //no callback verify context
	}

	//no error before verification
	cbCtx->error = J_SSL_NO_ERROR;
	
	//don't accept any error other than ASN_NO_SIGNER_E (no CA found error)
	if(store->error != ASN_NO_SIGNER_E){
		cbCtx->error = LLNET_SSL_TranslateReturnCode(store->error);
		return 0; 
	}
	
#ifdef SESSION_CERTS
	ssl = cbCtx->ssl;
	ctx = ssl->ctx;
	chainLength = ssl->session.chain.count;
	//check chain depth
	if(store->error_depth > chainLength){
		//max chain depth exceeded
		cbCtx->error = J_MAX_CHAIN_ERROR;
		return 0;
	}
	
	#ifdef WOLFSSL_SMALL_STACK
    DecodedCert* dCert = (DecodedCert*)XMALLOC(sizeof(DecodedCert), NULL, DYNAMIC_TYPE_TMP_BUFFER);
    if (dCert == NULL){
				cbCtx->error = J_MEMORY_ERROR; //out of memory
				return 0; 
		}        
	#else
    DecodedCert            dCert[1];
	#endif
		
	/*check if at least one of certificates in the chain is trusted.
		Start at the peer certificate which is at index 0 */
	int32_t ret = 0;
	int32_t trustedCertIndex = 0;
	Signer * ca = NULL;
	while(trustedCertIndex < chainLength && ca == NULL){
			byte* certBuffer = ssl->session.chain.certs[trustedCertIndex].buffer;
			int certLength = ssl->session.chain.certs[trustedCertIndex].length;
			InitDecodedCert(dCert, certBuffer, certLength, ssl->heap);
			//parse without cert verification
			ret = ParseCertRelative(dCert, CA_TYPE, 0, ctx->cm); 
			if(ret == 0){
					ca = LLNET_SSL_VERIFY_getTrustedCA(ctx, dCert);
					FreeDecodedCert(dCert);
			}
			if(ca == NULL){
				trustedCertIndex++;
			}		
	}
	
	if(ca == NULL){
		//no trusted certificate found
		cbCtx->error = J_NO_TRUSTED_CERT;	
	}
	
	//the peer certificate is trusted, verify is ok
	if(trustedCertIndex  == 0 && ca != NULL){
		cbCtx->error = J_SSL_NO_ERROR;
	}
	
	//the peer certificate is not trusted but an intermediate or the root certificate is trusted.
	//Validate the partial chain [0.. trustedCertIndex-1] by checking the basic constrains, name constraints, name chaining and signatures
	if(ca != NULL){
		
		Signer signer[1];
		signer->keyOID = ca->keyOID;
		signer->pubKeySize = ca->pubKeySize;
		signer->publicKey = ca->publicKey;
		signer->name = ca->name;
		signer->nameLen = ca->nameLen;
		#ifndef IGNORE_NAME_CONSTRAINTS
			signer->permittedNames = ca->permittedNames;
			signer->excludedNames  = ca->excludedNames;
		#endif
		#ifndef NO_SKID
			XMEMCPY(signer->subjectKeyIdHash , ca->subjectKeyIdHash, SHA_DIGEST_SIZE);
		#endif
		XMEMCPY(signer->subjectNameHash, ca->subjectNameHash, SHA_DIGEST_SIZE);
		signer->keyUsage  = ca->keyUsage ;
		
		//validate top down, start at trustedCertIndex-1
		for(int i = trustedCertIndex-1; i >=0 && cbCtx->error == J_SSL_NO_ERROR; i--){
			byte* certBuffer = ssl->session.chain.certs[i].buffer;
			int certLength = ssl->session.chain.certs[i].length;
			InitDecodedCert(dCert, certBuffer, certLength, ssl->heap);
			
			//parse with verification (only cert date validity will be checked)
			ret = ParseCertRelative(dCert, CA_TYPE, 1, ctx->cm);
			if(ret == 0){
					cbCtx->error = LLNET_SSL_VERIFY_validateCertChain(signer, dCert, (i!=0));
					if(cbCtx->error == J_SSL_NO_ERROR){
							//don't free the trusted CA (original trusted CA in ca table)
							if(i != trustedCertIndex-1){
									LLNET_SSL_VERIFY_freeSignerContent(signer, ctx->heap);
							}
							
							if(i != 0){
									LLNET_SSL_VERIFY_decodedCertAsSigner(signer, dCert);
									//set to NULL to avoid freeing during the FreeDecodedCert() function call.
									//This infos are used as signer infos for the next certificate validation in the chain
									//and it will be freed when LLNET_SSL_VERIFY_freeSignerContent() is called
									dCert->publicKey = NULL;
									dCert->subjectCN = NULL;
									dCert->permittedNames = NULL;
									dCert->excludedNames = NULL;
							}

							FreeDecodedCert(dCert);
					}else{
							//don't free the trusted CA
							if(i != trustedCertIndex-1){
									LLNET_SSL_VERIFY_freeSignerContent(signer, ctx->heap);
							}
							FreeDecodedCert(dCert);
					}
			}else{
					cbCtx->error = LLNET_SSL_TranslateReturnCode(ret);
					//don't free the trusted CA
					if(i != trustedCertIndex-1){
							LLNET_SSL_VERIFY_freeSignerContent(signer, ctx->heap);
					}
			}
		}
	}
	#ifdef WOLFSSL_SMALL_STACK
		XFREE(dCert, cm->heap, DYNAMIC_TYPE_CERT);
	#endif
#else
	cbCtx->error = LLNET_SSL_TranslateReturnCode(store->error);
#endif	/* SESSION_CERTS */

	return cbCtx->error == J_SSL_NO_ERROR? 1 : 0;
}

/* Validate the given certificate by verifying basic constraints, name constraints, name chaining and signature*/
int32_t LLNET_SSL_VERIFY_validateCertChain(Signer* issuer, DecodedCert* cert, uint8_t intermediateCert)
{
	//check basic constraints if intermediate cert
	if(intermediateCert && !cert->isCA){
		//intermediate cert must be a CA
		return J_NOT_CA_CERT; 
	}
	
	//check name chaining
	if(XMEMCMP(issuer->subjectNameHash, cert->issuerHash, SHA_DIGEST_SIZE) != 0){
		return J_NAME_CHAINING_ERROR; //name chaining error
	}
	
	//confirm signature
	if (!ConfirmSignature(cert->source + cert->certBegin,
							cert->sigIndex - cert->certBegin,
					issuer->publicKey, issuer->pubKeySize, issuer->keyOID,
					cert->signature, cert->sigLength, cert->signatureOID,
					cert->heap)) {
			WOLFSSL_MSG("Confirm signature failed");
			return J_ASN_SIG_CONFIRM_ERROR;
	}
	#ifndef IGNORE_NAME_CONSTRAINTS
	//check that this cert's name is permitted by the signer's name constraints
	if (!ConfirmNameConstraints(issuer, cert)) {
			WOLFSSL_MSG("Confirm name constraint failed");
			return J_ASN_NAME_INVALID_ERROR;
	}
#endif /* IGNORE_NAME_CONSTRAINTS */
	return J_SSL_NO_ERROR;
}

/* Turn cert as signer, just get signer infos*/
void LLNET_SSL_VERIFY_decodedCertAsSigner(Signer* issuer, DecodedCert* dCert)
{
	issuer->keyOID  = dCert->keyOID;
	issuer->pubKeySize = dCert->pubKeySize;
	issuer->publicKey = dCert->publicKey;
	issuer->name = dCert->subjectCN;
	issuer->nameLen = dCert->subjectCNLen;
	#ifndef IGNORE_NAME_CONSTRAINTS
	issuer->permittedNames = dCert->permittedNames;
	issuer->excludedNames  = dCert->excludedNames;
	#endif
	#ifndef NO_SKID
	XMEMCPY(issuer->subjectKeyIdHash , dCert->extSubjKeyId, SHA_DIGEST_SIZE);
	#endif
	XMEMCPY(issuer->subjectNameHash, dCert->subjectHash, SHA_DIGEST_SIZE);
	issuer->keyUsage  = dCert->extKeyUsage ;
}

/*free signer content*/
void LLNET_SSL_VERIFY_freeSignerContent(Signer* issuer, void* heap)
{
    XFREE(issuer->name, heap, DYNAMIC_TYPE_SUBJECT_CN);
    XFREE(issuer->publicKey, heap, DYNAMIC_TYPE_PUBLIC_KEY);
    #ifndef IGNORE_NAME_CONSTRAINTS
        if (issuer->permittedNames)
            FreeNameSubtrees(issuer->permittedNames, heap);
        if (issuer->excludedNames)
            FreeNameSubtrees(issuer->excludedNames, heap);
    #endif

}

#ifdef __cplusplus
	}
#endif
