/*
 *    C
 *
 *    Copyright 2016-2019 MicroEJ Corp. All rights reserved.
 *    MicroEJ Corp. PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */

/* Includes ------------------------------------------------------------------*/
#include "fsl_i2c.h"
#include "fsl_ft5406.h"
#include "fsl_gpio.h"
#include "fsl_inputmux.h"
#include "microej.h"
#include "io_task.h"
#include "touch_helper.h"
#include "FreeRTOS.h"
#include "task.h"
#include "semphr.h"
#include "board.h"

/* Definitions */
#define I2C_MASTER_BASE (I2C2_BASE)
#define I2C_MASTER_CLOCK_FREQUENCY (12000000)
#define I2C_MASTER ((I2C_Type *)I2C_MASTER_BASE)
#define I2C_MASTER_SLAVE_ADDR_7BIT 0x7EU
#define I2C_BAUDRATE 100000U

/* Task delay (in ms) between detecting a multiple pressed events */
#define TOUCH_DELAY 20

/* Task stack size */
#define TOUCH_STACK_SIZE  (128)

/* Global --------------------------------------------------------------------*/
static ft5406_handle_t touch_handle;
static SemaphoreHandle_t TS_Event;

/* Private API ---------------------------------------------------------------*/
/**
 * Retrieves the touch state and coordinates
 */
static void TOUCH_MANAGER_read(void)
{
    while (1)
    {
        touch_event_t touch_event;
	    int fingerX, fingerY;

        /* Read touch state and coordinates */
        if (FT5406_GetSingleTouch(&touch_handle, &touch_event, &fingerX, &fingerY) == kStatus_Success)
        {
            /* Check touch state */
            if ((touch_event == kTouch_Down) || (touch_event == kTouch_Contact))
            {
                /* Press event */
			    TOUCH_HELPER_pressed(fingerY, fingerX);

                /* Sleep some time before checking touch state again */
                vTaskDelay(TOUCH_DELAY);
            }
            else
            {
                /* Release event */
                TOUCH_HELPER_released();
                return;
            }
        }
        else
        {
            /* Invalid event */
            TOUCH_HELPER_released();
            return;
        }
    }
}

/**
 * Touch thread routine
 */
static void TOUCH_MANAGER_task(void *pvParameters)
{
    while (1)
    {
        /* Suspend ourselves */
        xSemaphoreTake(TS_Event, portMAX_DELAY);

        /* We have been woken up, lets work ! */
        TOUCH_MANAGER_read();

        /* Re-enable the touch interrupt */
        BOARD_EnableTSInterrupt();
    }
}

/**
 * Touch thread routine
 */
static void TOUCH_MANAGER_int_callback(bool pressed)
{
    portBASE_TYPE xHigherPriorityTaskWoken = pdFALSE;

    /* Disable the touch interrupt until after treatment */
    BOARD_DisableTSInterrupt();

    /* Set event */
    xSemaphoreGiveFromISR(TS_Event, &xHigherPriorityTaskWoken);

    /* Switch tasks if necessary */
    if (xHigherPriorityTaskWoken != pdFALSE)
    {
        portEND_SWITCHING_ISR(xHigherPriorityTaskWoken);
    }
}

static status_t I2C_Init(void)
{
    i2c_master_config_t masterConfig;

    I2C_MasterGetDefaultConfig(&masterConfig);

    /* Change the default baudrate configuration */
    masterConfig.baudRate_Bps = I2C_BAUDRATE;

    /* Initialize the I2C master peripheral */
    I2C_MasterInit(I2C_MASTER, &masterConfig, I2C_MASTER_CLOCK_FREQUENCY);

    return kStatus_Success;
}

/* Public functions ----------------------------------------------------------*/

void TOUCH_MANAGER_initialize(void)
{
    gpio_pin_config_t pin_config_rst = {
        kGPIO_DigitalOutput, 0,
    };

    /* Create a binary semaphore */
    vSemaphoreCreateBinary(TS_Event);

    /* Force the event to the non-signaled state */
    xSemaphoreTake(TS_Event, 0);

    /* Attach 12 MHz clock to FLEXCOMM2 (I2C master for touch controller) */
    CLOCK_AttachClk(kFRO12M_to_FLEXCOMM2);
    CLOCK_EnableClock(kCLOCK_Gpio2);

    /* Initialize I2C */
    if (I2C_Init() != kStatus_Success)
    {
        return;
    }

    /* Reset touch controller */
    GPIO_PinInit(GPIO, 2, 27, &pin_config_rst);
    GPIO_WritePinOutput(GPIO, 2, 27, 1);

    /* Initialize touch controller */
    if (FT5406_Init(&touch_handle, I2C_MASTER) != kStatus_Success)
    {
        return;
    }

    /* Register touch screen press and release interrupts */
    BOARD_RegisterTSInterruptCallback(TOUCH_MANAGER_int_callback);

    /* Re-enable the touch interrupt */
    BOARD_EnableTSInterrupt();

    /* Create the touch screen task */
    (void)xTaskCreate(TOUCH_MANAGER_task, "Touch screen task", TOUCH_STACK_SIZE, NULL, 13, NULL);
}

/* End of File */
