/*
 * Copyright (c) 2016, Freescale Semiconductor, Inc.
 * Copyright 2016-2017 NXP
 *
 * Redistribution and use in source and binary forms, with or without modification,
 * are permitted provided that the following conditions are met:
 *
 * o Redistributions of source code must retain the above copyright notice, this list
 *   of conditions and the following disclaimer.
 *
 * o Redistributions in binary form must reproduce the above copyright notice, this
 *   list of conditions and the following disclaimer in the documentation and/or
 *   other materials provided with the distribution.
 *
 * o Neither the name of the copyright holder nor the names of its
 *   contributors may be used to endorse or promote products derived from this
 *   software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 * ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "board.h"
#include <stdint.h>
#include "clock_config.h"
#include "fsl_common.h"
#include "fsl_debug_console.h"
#include "fsl_emc.h"
#include "fsl_sctimer.h"
#include "fsl_inputmux.h"
#include "pin_mux.h"
#include "spifilib_api.h"

/*******************************************************************************
 * Definitions
 ******************************************************************************/
/* The SDRAM timing. */
#define SDRAM_REFRESHPERIOD_NS (64 * 1000000 / 4096) /* 4096 rows/ 64ms */
#define SDRAM_TRP_NS (18u)
#define SDRAM_TRAS_NS (42u)
#define SDRAM_TSREX_NS (67u)
#define SDRAM_TAPR_NS (18u)
#define SDRAM_TWRDELT_NS (6u)
#define SDRAM_TRC_NS (60u)
#define SDRAM_RFC_NS (60u)
#define SDRAM_XSR_NS (67u)
#define SDRAM_RRD_NS (12u)
#define SDRAM_MRD_NCLK (2u)
#define SDRAM_RAS_NCLK (2u)
#define SDRAM_MODEREG_VALUE (0x23u)
#define SDRAM_DEV_MEMORYMAP (0x09u) /* 128Mbits (8M*16, 4banks, 12 rows, 9 columns)*/

/*******************************************************************************
 * Variables
 ******************************************************************************/
/* Variables needed for SPIFI initialization. */
char spifiMemBuffer[(sizeof(SPIFI_HANDLE_T) + sizeof(SPIFI_INFODATA_T) + 3) & ~0x03];
SPIFI_HANDLE_T * pSPIFIHandle;
uint32_t spifiHandleSize;
uint32_t sctimerTouchRiseEvent;
uint32_t sctimerTouchFallEvent;
uint32_t sctimerBacklightEvent;
static int_irq_touch_panel_type user_int_irq_touch_panel = NULL;

/*******************************************************************************
 * Code
 ******************************************************************************/
/* Initialize debug console. */
void BOARD_InitDebugConsole(void)
{
#if !defined(UNIT_TEST) && !defined(COREMARK_TEST)
    uint32_t usart_traces_switch = (uint32_t)(& _usart_traces_switch);
#else
    uint32_t usart_traces_switch = USART_TRACES_USB;
#endif
    status_t result;

    if (usart_traces_switch == USART_TRACES_USB) {
        /* Attach the 12 MHz clock to FLEXCOMM4 (debug console). */
        CLOCK_AttachClk(BOARD_DEBUG_USART0_CLK_ATTACH);

        /* Perform pin initialization. */
        BOARD_InitUsartPins();

        /* Reset the UART peripheral. */
        RESET_PeripheralReset(BOARD_DEBUG_USART0_RST);

        result = DbgConsole_Init(BOARD_DEBUG_USART0_BASEADDR, BOARD_DEBUG_UART_BAUDRATE, DEBUG_CONSOLE_DEVICE_TYPE_FLEXCOMM,
                BOARD_DEBUG_USART0_CLK_FREQ);
    } else if (usart_traces_switch == USART_TRACES_EXT) {
        /* Attach the 12 MHz clock to FLEXCOMM4 (debug console). */
        CLOCK_AttachClk(BOARD_DEBUG_USART4_CLK_ATTACH);

        /* Perform pin initialization. */
        BOARD_InitUsartPins();

        /* Reset the UART peripheral. */
        RESET_PeripheralReset(BOARD_DEBUG_USART4_RST);

        result = DbgConsole_Init(BOARD_DEBUG_USART4_BASEADDR, BOARD_DEBUG_UART_BAUDRATE, DEBUG_CONSOLE_DEVICE_TYPE_FLEXCOMM,
                BOARD_DEBUG_USART4_CLK_FREQ);
    }

    /* Check result. */
    assert(kStatus_Success == result);
}

void BOARD_SCTimerCallback(void)
{
    if (SCTIMER_GetStatusFlags(SCT0) & (1U << sctimerTouchRiseEvent))
    {
        SCTIMER_ClearStatusFlags(SCT0, 1U << sctimerTouchRiseEvent);
        if (user_int_irq_touch_panel)
        {
            user_int_irq_touch_panel(true);
        }
    }
    else if (SCTIMER_GetStatusFlags(SCT0) & (1U << sctimerTouchFallEvent))
    {
        SCTIMER_ClearStatusFlags(SCT0, 1U << sctimerTouchFallEvent);
        if (user_int_irq_touch_panel)
        {
            user_int_irq_touch_panel(false);
        }
    }
}

/* Initialize the LCD. */
void BOARD_InitLCD(void)
{
    sctimer_pwm_signal_param_t pwmParam;
    sctimer_config_t config;

    /* Configure LCD clock. */
    CLOCK_AttachClk(kMCLK_to_LCD_CLK);
    CLOCK_SetClkDiv(kCLOCK_DivLcdClk, 1, true);

    /* Perform LCD pin initialization. */
    BOARD_InitLcdPins();

    /* Perform I2C pin initialization. */
    BOARD_InitI2CPins();

    /* Perform Touch pin initialization. */
    BOARD_InitTouchPins();

    /* Configure SCTimer clock. */
    CLOCK_AttachClk(kMCLK_to_SCT_CLK);
    CLOCK_SetClkDiv(kCLOCK_DivSctClk, 2, true);

    /* Initialize the SCTimer module. */
    SCTIMER_GetDefaultConfig(&config);
    SCTIMER_Init(SCT0, &config);

    /* Initialize the Inputmux module. */
    INPUTMUX_Init(INPUTMUX);

    /* Configure the Inputmux for the SCTimer input 1 from external pin. */
    INPUTMUX_AttachSignal(INPUTMUX, 1U, kINPUTMUX_SctGpi1ToSct0);

    /* Configure PWM used for back light. */
    pwmParam.output = kSCTIMER_Out_5;
    pwmParam.level = kSCTIMER_HighTrue;
    pwmParam.dutyCyclePercent = 100U;
    SCTIMER_SetupPwm(SCT0, &pwmParam, kSCTIMER_CenterAlignedPwm, 1000U, CLOCK_GetFreq(kCLOCK_Sct), &sctimerBacklightEvent);

    /* Configure SCTimers used for touch interrupts (falling and rising events). */
    SCTIMER_CreateAndScheduleEvent(SCT0, kSCTIMER_InputFallEvent, 0U, kSCTIMER_Input_1, kSCTIMER_Counter_L, &sctimerTouchFallEvent);
    SCTIMER_CreateAndScheduleEvent(SCT0, kSCTIMER_InputRiseEvent, 0U, kSCTIMER_Input_1, kSCTIMER_Counter_L, &sctimerTouchRiseEvent);

    /* Set SCTimer callbacks */
    SCTIMER_SetCallback(SCT0, BOARD_SCTimerCallback, sctimerTouchFallEvent);
    SCTIMER_SetCallback(SCT0, BOARD_SCTimerCallback, sctimerTouchRiseEvent);

    /* Enable the SCTimer interrupt in NVIC. */
    NVIC_SetPriority(SCT0_IRQn, 7);
    NVIC_EnableIRQ(SCT0_IRQn);

    /* Start the SCTimer */
    SCTIMER_StartTimer(SCT0, kSCTIMER_Counter_L);
}

/* Enables the touch screen interrupt. */
void BOARD_EnableTSInterrupt(void)
{
    /* Enable SCTimer interrupts. */
    SCTIMER_EnableInterrupts(SCT0, (1U << sctimerTouchFallEvent) | (1U << sctimerTouchRiseEvent));
}

/* Disables the touch screen interrupt. */
void BOARD_DisableTSInterrupt(void)
{
    /* Disable SCTimer interrupts. */
    SCTIMER_DisableInterrupts(SCT0, (1U << sctimerTouchFallEvent) | (1U << sctimerTouchRiseEvent));
}

/* Registers the touch screen interrupt callback. */
void BOARD_RegisterTSInterruptCallback(int_irq_touch_panel_type callback)
{
    user_int_irq_touch_panel = callback;
}

/* Initialize the external memory. */
void BOARD_InitSDRAM(void)
{
    emc_basic_config_t basicConfig;
    emc_dynamic_timing_config_t dynTiming;
    emc_dynamic_chip_config_t dynChipConfig;

    /* Perform pin initialization. */
    BOARD_InitSdramPins();
    /* Basic configuration. */
    basicConfig.endian = kEMC_LittleEndian;
    basicConfig.fbClkSrc = kEMC_IntloopbackEmcclk;
    /* EMC Clock = CPU FREQ/2 here can fit CPU freq from 12M ~ 180M.
     * If you change the divide to 0 and EMC clock is larger than 100M
     * please take refer to emc.dox to adjust EMC clock delay.
     */
    basicConfig.emcClkDiv = 1;
    /* Dynamic memory timing configuration. */
    dynTiming.readConfig = kEMC_Cmddelay;
    dynTiming.refreshPeriod_Nanosec = SDRAM_REFRESHPERIOD_NS;
    dynTiming.tRp_Ns = SDRAM_TRP_NS;
    dynTiming.tRas_Ns = SDRAM_TRAS_NS;
    dynTiming.tSrex_Ns = SDRAM_TSREX_NS;
    dynTiming.tApr_Ns = SDRAM_TAPR_NS;
    dynTiming.tWr_Ns = (1000000000 / CLOCK_GetFreq(kCLOCK_EMC) + SDRAM_TWRDELT_NS); /* one clk + 6ns */
    dynTiming.tDal_Ns = dynTiming.tWr_Ns + dynTiming.tRp_Ns;
    dynTiming.tRc_Ns = SDRAM_TRC_NS;
    dynTiming.tRfc_Ns = SDRAM_RFC_NS;
    dynTiming.tXsr_Ns = SDRAM_XSR_NS;
    dynTiming.tRrd_Ns = SDRAM_RRD_NS;
    dynTiming.tMrd_Nclk = SDRAM_MRD_NCLK;
    /* Dynamic memory chip specific configuration: Chip 0 - MTL48LC8M16A2B4-6A */
    dynChipConfig.chipIndex = 0;
    dynChipConfig.dynamicDevice = kEMC_Sdram;
    dynChipConfig.rAS_Nclk = SDRAM_RAS_NCLK;
    dynChipConfig.sdramModeReg = SDRAM_MODEREG_VALUE;
    dynChipConfig.sdramExtModeReg = 0; /* it has no use for normal sdram */
    dynChipConfig.devAddrMap = SDRAM_DEV_MEMORYMAP;
    /* EMC Basic configuration. */
    EMC_Init(EMC, &basicConfig);
    /* EMC Dynamic memory configuration. */
    EMC_DynamicMemInit(EMC, &dynTiming, &dynChipConfig, 1);
}

#if SPIFI_ENABLE_RECOVERY_FEATURE
/*!
 * @brief Board specific toggle of the SPIFI CS pin
 */
static void BOARD_ToggleSpifiClk(uint8_t num_clocks)
{
    uint8_t index;

    /* Output the required number of clocks */
    for (index = 0; index < num_clocks; index++)
    {
        /* Drive SPIFI clk HIGH */
        GPIO_WritePinOutput(GPIO, 0, 26, 1);
        __ASM("NOP");

        /* Drive SPIFI clk LOW */
        GPIO_WritePinOutput(GPIO, 0, 26, 0);
        __ASM("NOP");
    }
}

/*!
 * @brief Board specific toggle of the SPIFI CS pin
 */
static void BOARD_ToggleSpifiCs(uint8_t num_clocks)
{
    /* The S# deselect time of minimum 50ns must be respected (at maximum frequency of 180MHz, the below loop takes 55ns) */
    BOARD_ToggleSpifiClk(10);

    /* Drive S# LOW */
    GPIO_WritePinOutput(GPIO, 0, 23, 0);

    /* Output the required number of clocks */
    BOARD_ToggleSpifiClk(num_clocks);

    /* Drive S# HIGH */
    GPIO_WritePinOutput(GPIO, 0, 23, 1);
}

/*!
 * @brief Board specific recovery of the SPIFI memory
 */
static void BOARD_PerformSpifiRecovery(uint8_t mode)
{
    gpio_pin_config_t gpioConfig = {kGPIO_DigitalOutput, 0};

    /* Enable GPIO0 clock */
    CLOCK_EnableClock(kCLOCK_Gpio0);

    /* SPIFI pin configuration (re-configure to GPIO mode). */
    BOARD_ReconfigureSpifiPinsForRecovery();

    /* GPIO config (set all GPIOs as digital output, low level) */
    GPIO_PinInit(GPIO, 0, 23, &gpioConfig);
    GPIO_PinInit(GPIO, 0, 24, &gpioConfig);
    GPIO_PinInit(GPIO, 0, 26, &gpioConfig);
    GPIO_PinInit(GPIO, 0, 27, &gpioConfig);

    /* Set DQ0 (PAD DATA) and DQ3 (PAD HOLD) HIGH */
    GPIO_WritePinOutput(GPIO, 0, 24, 1);
    GPIO_WritePinOutput(GPIO, 0, 27, 1);

    /* Perform the first step - Recovery */
    BOARD_ToggleSpifiCs(7);
    BOARD_ToggleSpifiCs(9);
    BOARD_ToggleSpifiCs(13);
    BOARD_ToggleSpifiCs(17);
    BOARD_ToggleSpifiCs(25);
    BOARD_ToggleSpifiCs(33);

    /* Perform the second step - mode dependent */
    if (mode == SPIFI_POWER_LOSS_RECOVERY) {
        /* Power Loss Recovery */
        BOARD_ToggleSpifiCs(8);
    }
    else if (mode == SPIFI_INTERFACE_RESCUE) {
        /* Interface Rescue */
        BOARD_ToggleSpifiCs(16);
    }

    /* SPIFI pin configuration (re-configure to SPIFI mode). */
    BOARD_InitSpifiPins();

    /* Disable GPIO0 clock */
    CLOCK_DisableClock(kCLOCK_Gpio0);
}
#endif

/*!
 * @brief Board specific initialization of the SPIFI
 */
void BOARD_InitSPIFI(void)
{
    /* Disable SPIFI clock. */
    CLOCK_DisableClock(kCLOCK_Spifi);

    /* Attach the 12 MHz clock to SPI3. */
    CLOCK_AttachClk(kFRO12M_to_FLEXCOMM9);

    /* Perform pin initialization. */
    BOARD_InitSpifiPins();

    /* Configure SPIFI clock. */
    CLOCK_AttachClk(kMAIN_CLK_to_SPIFI_CLK);
    CLOCK_SetClkDiv(kCLOCK_DivSpifiClk, CLOCK_GetFreq(kCLOCK_SpiFi) / SPIFI_MAX_FREQUENCY + 1, true);

    /* Enable SPIFI clock. */
    CLOCK_EnableClock(kCLOCK_Spifi);

    /* Perform SPIFI initialization. */
    spifiInit(SPIFI0_BASE, true);
    spifiRegisterFamily(spifi_REG_FAMILY_CommonCommandSet);
    spifiHandleSize = spifiGetHandleMemSize(SPIFI0_BASE);

    /* Detect initialization errors. */
    if (spifiHandleSize == 0)
    {
#if SPIFI_ENABLE_RECOVERY_FEATURE
        /* Perform SPIFI power-loss recovery */
        BOARD_PerformSpifiRecovery(SPIFI_POWER_LOSS_RECOVERY);

        /* Try again to initialize */
        spifiHandleSize = spifiGetHandleMemSize(SPIFI0_BASE);

        /* If it fails also now, then exit with error */
        if (spifiHandleSize == 0)
        {
            /* SPIFILIB cannot detect SPI Flash device. */
            CLOCK_DisableClock(kCLOCK_Spifi);
            return;
        }
#else
        /* SPIFILIB cannot detect SPI Flash device. */
        CLOCK_DisableClock(kCLOCK_Spifi);
        return;
#endif
    }

    /* Detect initialization errors. */
    if (spifiHandleSize > sizeof(spifiMemBuffer))
    {
        /* Handle mem buffer requires more space or dynamic allocation. May be due to updated SPIFILIB library. */
        CLOCK_DisableClock(kCLOCK_Spifi);
        return;
    }

    /* Init device. */
    pSPIFIHandle = spifiInitDevice(spifiMemBuffer, spifiHandleSize, SPIFI0_BASE, FSL_FEATURE_SPIFI_START_ADDR);
    if (!pSPIFIHandle)
    {
        /* Failed to initialize device. */
        CLOCK_DisableClock(kCLOCK_Spifi);
        return;
    }

    /* Enable quad mode. */
    if (spifiDevSetOpts(pSPIFIHandle, SPIFI_OPT_USE_QUAD, 1) != SPIFI_ERR_NONE)
    {
        /* Not supported option. */
        CLOCK_DisableClock(kCLOCK_Spifi);
        return;
    }

    /* Enter memory mode. */
    if (spifiDevSetMemMode(pSPIFIHandle, 1) != SPIFI_ERR_NONE)
    {
        /* Not supported option. */
        CLOCK_DisableClock(kCLOCK_Spifi);
        return;
    }
}

/*!
 * @brief Board specific initialization of the ENET
 */
void BOARD_InitENET(void)
{
    /* Perform pin initialization. */
    BOARD_InitEnetPins();
}

/*!
 * @brief Board specific initialization of the SDIF
 */
void BOARD_InitSDIF(void)
{
    /* Configure SDIF clock. */
    CLOCK_AttachClk(kMCLK_to_SDIO_CLK);
    CLOCK_SetClkDiv(kCLOCK_DivSdioClk, CLOCK_GetFreq(kCLOCK_SDio) / SDIF_MAX_FREQUENCY + 1U, true);

    /* Perform pin initialization. */
    BOARD_InitSdifPins();

    /* Set the SDIO interrupt priority in NVIC. */
    NVIC_SetPriority(SDIO_IRQn, 5);
}
